--
-- Created by IntelliJ IDEA.
-- User: insai
-- Date: 7/7/2016
-- Time: 9:55 PM
-- To change this template use File | Settings | File Templates.
--
local L = LibStub("AceLocale-3.0"):GetLocale("KindredUtils", true);
local LA = LibStub:GetLibrary("LibArtifactData-1.0")

KindredUtils = LibStub("AceAddon-3.0"):NewAddon("KindredUtils", "AceConsole-3.0", "AceEvent-3.0", "AceTimer-3.0", "AceComm-3.0", "AceSerializer-3.0", "AceDB-3.0")

KUTIL_ColorIndex = {0,10,15,20,25,30,35}
KUTIL_Colors = {
    -- White base color
    [0] =       {['r']=255,     ['g']=255,      ['b']=255,      ['rgb']='FFFFFF',   ['p']=0,},
    -- Yellow for > 10 traits
    [10] =     {['r']=255,     ['g']=255,      ['b']=0,        ['rgb']='FFFF00',   ['p']=0,},
    -- Green for > 15 traits
    [15] =     {['r']=0,       ['g']=255,      ['b']=0,        ['rgb']='00FF00',   ['p']=10,},
    -- Teal for > 20 traits
    [20] =     {['r']=0,       ['g']=255,      ['b']=255,      ['rgb']='00FFFF',   ['p']=15,},
    -- Blue for > 25 traits
    [25] =     {['r']=0,       ['g']=102,      ['b']=255,      ['rgb']='0066ff',   ['p']=20,},
    -- Purple for > 30 traits
    [30] =     {['r']=255,     ['g']=0,        ['b']=255,      ['rgb']='FF00FF',   ['p']=25,},
    -- Red for >= 35 traits
    [35] =    {['r']=255,     ['g']=0,        ['b']=0,        ['rgb']='FF0000',   ['p']=30,},
}
KUTIL_grayScore = 9
KUTIL_prefix = "KUTIL_"
KUTIL_prefixArtifacts = "artifacts"
KUTIL_prefixRequestArtifacts = "rArtifacts"

function KindredUtils:UpdateAndBroadcastArtifacts()
    self:UpdateArtifacts()
    self:BroadcastArtifacts()
end

function KindredUtils:ForceRequestArtifacts()
    self:RequestArtifactInfoFromPartyMembers(true)
end

function KindredUtils:UpdateArtifacts()
    if self.MyUnitGUID == nil or InCombatLockdown() then return false end
    local myCurrentArtifacts = self.db.global.artifactCache[self.MyUnitGUID]
    if myCurrentArtifacts == nil then
        self:Print("No current artifacts tracked, getting all infor and storing it.")
        myCurrentArtifacts = LA.GetAllArtifactsInfo()
    else
        self:Print("Have pre-existing artifactss. Updating list")
        local toBeCached = LA.GetAllArtifactsInfo()
        for k, v in pairs(toBeCached) do
            self:Print("Adding / Updating key " .. k)
            if k ~= 'knowledgeLevel' and k ~= 'knowledgeMultiplier' then
                myCurrentArtifacts[k] = v
            end
        end
    end
    self.db.global.artifactCache[self.MyUnitGUID] = myCurrentArtifacts
end

function KindredUtils:GetCurrentArtifactInfo()
    local myCacheEntry = self.db.global.artifactCache[self.MyUnitGUID]
    if myCacheEntry then
        local myArtifactInfo = self.db.global.artifactCache[self.MyUnitGUID][LA.GetActiveArtifactID()]
        if myArtifactInfo then
            local name = myArtifactInfo['name']
            local traits = myArtifactInfo['numRanksPurchased']
            local knowledge = {}
            knowledge['level'] = myArtifactInfo['knowledgeLevel']
            knowledge['multiplier'] = myArtifactInfo['knowledgeMultiplier']
            return name, traits, knowledge
        else
            return false
        end
    end
end

function KindredUtils:GetArtifactInfo(id)
    local myArtifactInfo = self.db.global.artifactCache[self.MyUnitGUID][id]
    local name = myArtifactInfo['name']
    local traits = myArtifactInfo['numRanksPurchased']
    return name, traits
end

function KindredUtils:DebugPrint()
    local name = UnitName("player")
    local guid = UnitGUID(name)
    self:Print("My Unit GUID: " .. guid)
    self:PrintR(self.db.global.artifactCache)
end

function KindredUtils:ClearCache()
    self.db.global.artifactCache = {}
end

function KindredUtils:GetArtifactWeapons(guid)
    if not self:IsGUID(guid, 'Player') then return false; end

    if self.db.global.artifactCache[guid] then
        weapons = {}
        knowledge = {}
        for k, v in pairs(self.db.global.artifactCache[guid]) do
            if k ~= 'knowledgeLevel' and k ~= 'knowledgeMultiplier' then
                table.insert(weapons, v)
            else
                knowledge[k] = v
            end
        end
        return weapons, knowledge
    else
        return {}, {}
    end
end

function KindredUtils:GetGoldTraitCount(traits)
    local goldCount = 0
    if traits ~= nil then
        for k, v in pairs(traits) do
            if v.isGold then goldCount = goldCount + 1 end
        end
    end
    return goldCount
end

--
-- Credit where credit is due - The following functions were borrowed from SimpleILevel by Scotepi
--

-- Used to hook the tooltip to avoid the full tooltip function
function KindredUtils:TooltipHook()
    local name, unit = GameTooltip:GetUnit();
    local guid = false;
    
    if unit then
        guid = UnitGUID(unit);
    elseif name then
        guid = self:NameToGUID(name);
    end
    if self:IsGUID(guid, 'Player') then
        self:ShowTooltip(guid, unit);
    end
end

function KindredUtils:NameToGUID(name, realm)
    if name then
        return UnitGUID(name)
    end
    return false;
end

function KindredUtils:IsGUID(guid, type)
    if not guid then return false end
    if not type then type = 'player' end

    local gType, gGuid = strsplit('-', guid, 2);

    if strlower(type) == strlower(gType) then 
        return true
    else 
        return false
    end
end

function KindredUtils:ShowTooltip(guid, unit)
    if InCombatLockdown() then return end
    
    if not guid then
        guid = UnitGUID("mouseover")
        if self:IsGUID(guid, "player") then
            unit = UnitId("mouseover")
        end
    end
    
    local weapons, knowledge = self:GetArtifactWeapons(guid)
    
    -- Artifact Weapons
    local noArtifactInfo = true
    if weapons then
        for k,v in pairs(weapons) do
            -- Build the tooltip text
            noArtifactInfo = false
            local traitsHex, r, g, b = self:ColorArtifactTraits(v.numRanksPurchased)
            local goldTraits = self:GetGoldTraitCount(v.traits)
            local textLeft = '|cFF216bff' .. v.name .. ':|r '
            local textRight = '|cFF' .. traitsHex .. ' ' .. v.numRanksPurchased .. '|r |cFFFFD700(' .. goldTraits .. ')|r |cFFFFFFFF' .. L.core.traitsUnlocked .. '|r'
            self:AddTooltipText(textLeft, textRight);
        end
    else
        noArtifactInfo = true
    end

    if noArtifactInfo then
        local textLeft = '|cFF216bff' .. L.core.artifactInformationUnavailable .. '|r '
        self:AddTooltipText(textLeft, '');
        self:RequestArtifactInfoFromPartyMembers(guid)
    end
    -- Artifact Knowledge
    if knowledge and knowledge['knowledgeLevel'] and knowledge['knowledgeMultiplier'] then
        local textLeft = '|cFF216bff' .. L.core.akLabel .. ':|r '
        local textRight = knowledge['knowledgeLevel'] .. ' (' .. (knowledge['knowledgeMultiplier'] - 1) * 100 .. '%)'
        self:AddTooltipText(textLeft, textRight)
    end
    return true;
end

-- Return the hex, r, g, b of a score
function KindredUtils:ColorArtifactTraits(traits)
    -- There are some missing items so gray
    if traits and traits <= KUTIL_grayScore then
        return self:RGBtoHex(0.5,0.5,0.5), 0.5,0.5,0.5;
    end
    
    -- Default to white
    local r,g,b = 1,1,1;
    
    local found = false;

    for i,maxScore in pairs(KUTIL_ColorIndex) do
        if traits < maxScore and not found then
            local colors = KUTIL_Colors[maxScore];
            local baseColors = KUTIL_Colors[colors.p];
            
            local steps = maxScore - colors.p;
            local scoreDiff = traits - colors.p;
            
            local diffR = (baseColors.r - colors.r) / 255;
            local diffG = (baseColors.g - colors.g) / 255;
            local diffB = (baseColors.b - colors.b) / 255;
            
            local diffStepR = diffR / steps;
            local diffStepG = diffG / steps;
            local diffStepB = diffB / steps;
            
            local scoreDiffR = scoreDiff * diffStepR;
            local scoreDiffG = scoreDiff * diffStepG;
            local scoreDiffB = scoreDiff * diffStepB;
            
            r = (baseColors.r / 255) - scoreDiffR;
            g = (baseColors.g / 255) - scoreDiffG;
            b = (baseColors.b / 255) - scoreDiffB;
            
            found = true;
        end
    end
    
    -- Nothing was found so max
    if not found then
        r = KUTIL_Colors[35].r;
        g = KUTIL_Colors[35].g;
        b = KUTIL_Colors[35].b;
    end
    
    local hex = self:RGBtoHex(r,g,b);
    return hex, r, g, b;
end

-- from http://www.wowpedia.org/RGBPercToHex
function KindredUtils:RGBtoHex(r, g, b)
    r = r <= 1 and r >= 0 and r or 0
    g = g <= 1 and g >= 0 and g or 0
    b = b <= 1 and b >= 0 and b or 0
    return string.format("%02x%02x%02x", r*255, g*255, b*255)
end

-- Add lines to the tooltip, testLeft must be the same
function KindredUtils:AddTooltipText(textLeft, textRight)
    
    -- Loop tooltip text to check if its alredy there
    local ttLines = GameTooltip:NumLines();
    local ttUpdated = false;
    
    for i = 1,ttLines do
        
        -- If the static text matches
        if _G["GameTooltipTextLeft"..i]:GetText() == textLeft then
            
            -- Update the text
            _G["GameTooltipTextLeft"..i]:SetText(textLeft);
            _G["GameTooltipTextRight"..i]:SetText(textRight);
            GameTooltip:Show();
            
            -- Remember that we have updated the tool tip so we won't again
            ttUpdated = true;
            break;
        end
    end
    
    -- Tooltip is new
    if not ttUpdated then
        GameTooltip:AddDoubleLine(textLeft, textRight);
        GameTooltip:Show();
    end
end

--
-- End Credited Code
--

function KindredUtils:SerializeArtifacts()
    if self.MyUnitGUID == nil then return end
    local ret = {}
    ret['GUID'] = self.MyUnitGUID
    ret['artifacts'] = self.db.global.artifactCache[self.MyUnitGUID]
    return self:Serialize(ret)
end

function KindredUtils:BroadcastArtifacts(force)
    self:UpdateArtifacts()
    -- Be kind, send low-priority messages, only send one every 15 seconds to avoid spamming
    -- Only send messages if you have one or more members of your group that are not from LFG/R
    if ((self.lastBroadcast == nil or time() - self.lastBroadcast > 15) and GetNumGroupMembers(LE_PARTY_CATEGORY_HOME) > 0) or force then
        local message = self:SerializeArtifacts()
        if message ~= nil then
            self:SendCommMessage(KUTIL_prefix .. KUTIL_prefixRequestArtifacts, message, "RAID", nil, "BULK")
            self.lastBroadcast = time()
        end
    end
end

function KindredUtils:RequestArtifactInfoFromPartyMembers(force, targetUnit)
    -- Be kind, don't spam requests -- 15 seconds between requests min
    -- Only send requests when you are in a party with atleast 1 person that is not from LFG/R
    -- Only send requests when the target GUID is for a player in your party
    if GetNumGroupMembers(LE_PARTY_CATEGORY_HOME) > 0 then
        local inPartyUnit = false
        if targetGUID and UnitInParty(targetUnit) then
            inPartyUnit = true
        end
        if (self.lastRequestTime == nill or time() - self.lastRequestTime > 15 and inPartyUnit) or force then
            self:SendCommMessage(KUTIL_prefix .. KUTIL_prefixRequestArtifacts, '', 'RAID', nil, 'BULK')
            self.lastRequestTime = time()
        end
    end
end

function KindredUtils:OnCommReceived(prefix, message, distribution, sender)
    if prefix == KUTIL_prefix .. KUTIL_prefixArtifacts then
        local success, payload = self:Deserialize(message)
        if success and payload.GUID then
            local thisPersonsArtifacts = {}
            if self.db.global.artifactCache[payload.GUID] then
                thisPersonsArtifacts = self.db.global.artifactCache[payload.GUID]
            end
            if payload.artifacts then
                self.db.global.artifactCache[payload.GUID] = payload.artifacts
            else
                self:Print("Bad payload received. No Artifact Info in message")
            end
        else
            self:Print("Bad payload received. No Character GUID in message")
        end
    elseif prefix == KUTIL_prefix .. KUTIL_prefixRequestArtifacts then
        self:Print("Received artifact info request from " .. sender)
        self:BroadcastArtifacts(true)
    end
end

function KindredUtils:RegisterPaperdoll()    
    self:Print("Registering Paperdoll Items")
    table.insert(PAPERDOLL_STATCATEGORIES[1].stats, 1, {
          stat = 'KUTIL_ak',
    });
        table.insert(PAPERDOLL_STATCATEGORIES[1].stats, 1, {
          stat = 'KUTIL_atu',
    });
    PAPERDOLL_STATINFO['KUTIL_ak'] = { updateFunc = function(...) KindredUtils:UpdatePaperDollFrameAk(...); end };
    PAPERDOLL_STATINFO['KUTIL_atu'] = { updateFunc = function(...) KindredUtils:UpdatePaperDollFrameAtu(...); end };
    self:Print("Done registering Paperdoll Items")
    self:PrintR(PAPERDOLL_STATCATEGORIES)
end

function KindredUtils:UnregisterPaperdoll()
    self:Print("Unregisering Paperdoll Items") 
    table.foreach(PAPERDOLL_STATCATEGORIES[1].stats, function(k, v)
        if v.stat == 'KUTIL_ak' then
            table.remove(PAPERDOLL_STATCATEGORIES[1].stats, k)
        elseif v.stat == 'KUTIL_atu' then
            table.remove(PAPERDOLL_STATCATEGORIES[1].stats, k)
        end
    end)
    PAPERDOLL_STATINFO['KUTIL_ak'] = { updateFunc = function(...) return false; end };
    PAPERDOLL_STATINFO['KUTIL_atu'] = { updateFunc = function(...) return false; end };
    self:Print("Done Unregisering Paperdoll Items")
end

function KindredUtils:UpdatePaperDollFrameAtu(statFrame, unit)
    local artifactName, artifactTraits = self:GetCurrentArtifactInfo()
    if artifactName and artifactTraits then
        local traitsHex = self:ColorArtifactTraits(artifactTraits)
        local formated = '|cFF' .. traitsHex .. artifactTraits .. '|r'
        
        PaperDollFrame_SetLabelAndText(statFrame, L.core.atuLabel, formated, false);
        statFrame.tooltip = HIGHLIGHT_FONT_COLOR_CODE..L.core.name..' - '..L.core.atuLabel..FONT_COLOR_CODE_CLOSE;
        statFrame.tooltip2 = L.core.atuDesc;
        statFrame:Show();
    end
end

function KindredUtils:UpdatePaperDollFrameAk(statFrame, unit)
    local artifactName, artifactTraits = self:GetCurrentArtifactInfo()
    local ak, akm = LA.GetArtifactKnowledge()
    if ak == -1 then ak = 'Unknown' end
    if akm == 0 then akm = 'Unknown Multiplier' else akm = (akm - 1) * 100  .. '%' end
    
    PaperDollFrame_SetLabelAndText(statFrame, L.core.akLabel, ak, false);
    statFrame.tooltip = HIGHLIGHT_FONT_COLOR_CODE..L.core.name..' - '..L.core.akLabel..FONT_COLOR_CODE_CLOSE;
    statFrame.tooltip2 = akm .. '\n\n' .. L.core.akDesc;
    statFrame:Show();
end

function KindredUtils:PrintR ( t )
    local print_r_cache={}
    local function sub_print_r(t,indent)
        if (print_r_cache[tostring(t)]) then
            self:Print(indent.."*"..tostring(t))
        else
            print_r_cache[tostring(t)]=true
            if (type(t)=="table") then
                for pos,val in pairs(t) do
                    if (type(val)=="table") then
                        self:Print(indent.."["..pos.."] => "..tostring(t).." {")
                        sub_print_r(val,indent..string.rep(" ",string.len(pos)+8))
                        self:Print(indent..string.rep(" ",string.len(pos)+6).."}")
                    elseif (type(val)=="string") then
                        self:Print(indent.."["..pos..'] => "'..val..'"')
                    else
                        self:Print(indent.."["..pos.."] => "..tostring(val))
                    end
                end
            else
                self:Print(indent..tostring(t))
            end
        end
    end
    if (type(t)=="table") then
        self:Print(tostring(t).." {")
        sub_print_r(t,"  ")
        self:Print("}")
    else
        sub_print_r(t,"  ")
    end
    self:Print()
end

function KindredUtils:OnInitialize()
    -- Code that runs when the addon is first loaded
    GameTooltip:HookScript("OnTooltipSetUnit", function(...) KindredUtils:TooltipHook(...); end);
    self.db = LibStub("AceDB-3.0"):New("KindredUtilsDB")
    self:RegisterComm(KUTIL_prefix .. KUTIL_prefixArtifacts)
    self:RegisterComm(KUTIL_prefix .. KUTIL_prefixRequestArtifacts)

    -- Add to Paperdoll
    self:RegisterPaperdoll();
end

function KindredUtils:OnEnable()
    -- Code that runs when the addon is enabled
    if self.db.global.artifactCache == nil then
        self.db.global.artifactCache = {}
    end 
    self.grayScore = 9
    if self.MyUnitGUID == nil then
        self.MyUnitGUID = self:NameToGUID(UnitName("player"))
    end

    -- Register Chat Commands
    self:RegisterChatCommand("kindredDebug", "DebugPrint")
    self:RegisterChatCommand("kindredClearCache", "ClearCache")
    self:RegisterChatCommand("kindredForceRequest", "ForceRequestArtifacts")
    self:RegisterChatCommand("kindredForceUpdate", "UpdateArtifacts")

    -- Send messages when party or raid changes
    self:RegisterEvent("PARTY_MEMBERS_CHANGED", "BroadcastArtifacts")
    self:RegisterEvent("RAID_ROSTER_UPDATE", "BroadcastArtifacts")

    -- Update artifacts and then send messages when certain events happen
    LA.RegisterCallback(self, "ARTIFACT_ADDED", "UpdateAndBroadcastArtifacts")
    LA.RegisterCallback(self, "ARTIFACT_ACTIVE_CHANGED", "UpdateAndBroadcastArtifacts")
    LA.RegisterCallback(self, "ARTIFACT_EQUIPPED_CHANGED", "UpdateAndBroadcastArtifacts")
    LA.RegisterCallback(self, "ARTIFACT_KNOWLEDGE_CHANGED", "UpdateAndBroadcastArtifacts")
    LA.RegisterCallback(self, "ARTIFACT_POWER_CHANGED", "UpdateAndBroadcastArtifacts")
    LA.RegisterCallback(self, "ARTIFACT_RELIC_CHANGED", "UpdateAndBroadcastArtifacts")
    LA.RegisterCallback(self, "ARTIFACT_TRAITS_CHANGED", "UpdateAndBroadcastArtifacts")

    self:UpdateArtifacts()
end

function KindredUtils:OnDisable()
    -- Code that runs when the addon is disabled
    -- Unregister Chat Commands
    self:UnregisterChatCommand("kindredDebug")
    self:UnregisterChatCommand("kindredClearCache")
    self:RegisterChatCommand("kindredForceRequest")
    self:UnregisterChatCommand("kindredForceUpdate")

    -- Unregister Events
    self:UnregisterEvent("PARTY_MEMBERS_CHANGED", "BroadcastArtifacts")
    self:UnregisterEvent("RAID_ROSTER_UPDATE", "BroadcastArtifacts")

    -- Unregister Callbacks
    LA.RegisterCallback(self, "ARTIFACT_ADDED")
    LA.RegisterCallback(self, "ARTIFACT_ACTIVE_CHANGED")
    LA.RegisterCallback(self, "ARTIFACT_EQUIPPED_CHANGED")
    LA.RegisterCallback(self, "ARTIFACT_KNOWLEDGE_CHANGED")
    LA.RegisterCallback(self, "ARTIFACT_POWER_CHANGED")
    LA.RegisterCallback(self, "ARTIFACT_RELIC_CHANGED")
    LA.RegisterCallback(self, "ARTIFACT_TRAITS_CHANGED")

    -- Unregister Paperdoll Handler
    self:UnregisterPaperdoll()
end
